/*
  UMainForm.cpp

  This file is part of LPCCISP.

  LPCCISP was developed by Synapse(Hiroshi Tanigawa) in 2014.  LPCCISP
  is originally distributed at "Synapse's history of making gadgets"
  <http://www3.big.or.jp/~schaft/hardware/>

  LPCCISP is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  LPCCISP is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with LPCCISP.  If not, see <http://www.gnu.org/licenses/>.
*/

//---------------------------------------------------------------------------

#include <vcl.h>
#include <winbase.h>
#include <stdio.h>
#include <stdlib.h>
#pragma hdrstop


#include "UMainForm.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;

static const int CPUfreq=12000;
static const int SectorSize=4096;
static const int RamBufAdr=0x10000200;
static const int RamBufSize=0x400;
static const int FlashSize=0x8000;


// UART ISP Return Codes
static const char CMD_SUCCESS                            []="0";
static const char INVALID_COMMAND                        []="1";
static const char SRC_ADDR_ERROR                         []="2";
static const char DST_ADDR_ERROR                         []="3";
static const char SRC_ADDR_NOT_MAPPED                    []="4";
static const char DST_ADDR_NOT_MAPPED                    []="5";
static const char COUNT_ERROR                            []="6";
static const char INVALID_SECTOR                         []="7";
static const char SECTOR_NOT_BLANK                       []="8";
static const char SECTOR_NOT_PREPARED_FOR_WRITE_OPERATION[]="9";
static const char COMPARE_ERROR                          []="10";
static const char BUSY                                   []="11";
static const char PARAM_ERROR                            []="12";
static const char ADDR_ERROR                             []="13";
static const char ADDR_NOT_MAPPED                        []="14";
static const char CMD_LOCKED                             []="15";
static const char INVALID_CODE                           []="16";
static const char INVALID_BAUD_RATE                      []="17";
static const char INVALID_STOP_BIT                       []="18";
static const char CODE_READ_PROTECTION_ENABLED           []="19";

static const char CancelMsg[]="tbV݂̏𒆒f܂";

static bool busy=false;
static bool CancelBtnPressed;

bool SendLine(HANDLE hCom, const char *str)
{
  DWORD cnt;

  if(!WriteFile(hCom,str,strlen(str),&cnt,NULL)||cnt!=strlen(str)) return false;
  if(!WriteFile(hCom,"\r\n",2,&cnt,NULL)||cnt!=2) return false;
  return true;
} // SendLine

bool ReceiveLine(HANDLE hCom, char *str, int len)
{
  int i;
  DWORD cnt;

  i=0;
  while(true) {
    if(!ReadFile(hCom,&str[i],1,&cnt,NULL)||cnt!=1) return false;
    if(str[i]=='\n') break;
    if((unsigned char)str[i]>=0x20) {
      if(++i>=len) {
        return false;
      } // if
    } // if
  } // while

  str[i]='\0';
  return true;
} // ReceiveLine

bool WaitForResponse(HANDLE hCom, const char *res)
{
  char str[100];

  return ReceiveLine(hCom,str,sizeof(str))&&strcmp(str,res)==0;
} // WaitForResponse

bool EnterIspMode(HANDLE hCom)
{
  const int RetryCnt=5;
  DWORD cnt;

  char str[100];

  for(int i=0; i<RetryCnt; i++) {
    // enable BOOT
    if(!EscapeCommFunction(hCom,SETRTS)) continue;

    // enable RESET
    if(!EscapeCommFunction(hCom,SETDTR)) continue;

    Sleep(50);

    // desable RESET
    if(!EscapeCommFunction(hCom,CLRDTR)) continue;

    Sleep(100);

    // disable BOOT
    if(!EscapeCommFunction(hCom,CLRRTS)) continue;

    PurgeComm(hCom, PURGE_RXABORT|PURGE_RXCLEAR);

    if(!WriteFile(hCom,"?",1,&cnt,NULL)||cnt!=1) continue;
    if(!WaitForResponse(hCom,"Synchronized")) continue;

    if(!SendLine(hCom,"Synchronized")) continue;
    if(!WaitForResponse(hCom,"Synchronized")) continue; // check echo back
    if(!WaitForResponse(hCom,"OK")) continue;

    sprintf(str,"%d",CPUfreq);
    if(!SendLine(hCom,str)) return false;
    if(!WaitForResponse(hCom,str)) continue; // check echo back
    if(!WaitForResponse(hCom,"OK")) continue;

    return true;
  } // for i

  return false;
} // EnterIspMode

bool KillEcho(HANDLE hCom)
{
  if(!SendLine(hCom,"A 0")) return false;
  if(!WaitForResponse(hCom,"A 0")) return false;
  return WaitForResponse(hCom,CMD_SUCCESS);
} // KillEcho

bool SetBaudRate(HANDLE hCom, int baud, int StopBit)
{
  char str[100];

  sprintf(str,"B %d %d",baud,StopBit);
  SendLine(hCom,str);
  return WaitForResponse(hCom,CMD_SUCCESS);
} // SetBaudRate

bool ReadPartId(HANDLE hCom, char *PartId, int len)
{
  if(!SendLine(hCom,"J")) return false;
  if(!WaitForResponse(hCom,CMD_SUCCESS)) return false;
  return ReceiveLine(hCom,PartId,len);
} // ReadPartId

bool unlock(HANDLE hCom)
{
  if(!SendLine(hCom,"U 23130")) return false;
  return WaitForResponse(hCom,CMD_SUCCESS);
} // unlock

bool EraseFlash(HANDLE hCom, int SectorNum)
{
  char str[100];

  // prepare sectors
  sprintf(str,"P 0 %d",SectorNum-1);
  if(!SendLine(hCom,str)) return false;
  if(!WaitForResponse(hCom,CMD_SUCCESS)) return false;

  // erase sectors
  sprintf(str,"E 0 %d",SectorNum-1);
  if(!SendLine(hCom,str)) return false;
  return WaitForResponse(hCom,CMD_SUCCESS);
} // EraseFlash

void UUencode(char *str, unsigned char *data, int len)
{
  int index;
  int data24;

  *(str++)=len+0x20;
  index=0;
  while(index<len) {
    data24=data[index]<<16;
    if(index+1<len) data24+=data[index+1]<<8;
    if(index+2<len) data24+=data[index+2]   ;

    for(int i=3; i>=0; i--) {
      unsigned char b=(data24>>(i*6))&0x3f;
      if(b==0) {
        b=0x60;
      } else {
        b+=0x20;
      } // if
      *(str++)=(char)b;
    } // for i

    index+=3;
  } // while
  *str='\0';
} // UUencode

bool WriteFlash(HANDLE hCom, FILE *f, int FileSize, void (*UpdateProgressBar)(int))
{
  unsigned adr;
  unsigned WriteSize;
  unsigned char WriteBuf[RamBufSize];
  char str[100];
  unsigned LineIndex;
  unsigned LineCnt;
  unsigned sum;
  unsigned LineSize;
  unsigned sector;

  adr=0;
  while(adr<(unsigned)FileSize) {
    (*UpdateProgressBar)(adr*100/FileSize);
    if (CancelBtnPressed) return false;

    WriteSize=FileSize-adr;
    if(WriteSize>RamBufSize) WriteSize=RamBufSize;

    // read data from file
    if(fread(WriteBuf,1,WriteSize,f)!=WriteSize) return false;

    // send W command
    sprintf(str,"W %d %d",RamBufAdr,WriteSize);
    if(!SendLine(hCom,str)) return false;
    if(!WaitForResponse(hCom,CMD_SUCCESS)) return false;

    // send data to MPU's RAM
    LineIndex=LineCnt=sum=0;
    while(LineIndex<WriteSize) {
      LineSize=WriteSize-LineIndex;
      if(LineSize>45) LineSize=45; // Maxmum number of bytes that can be sent in one line is limited to 45.
      UUencode(str,WriteBuf+LineIndex,LineSize);
      for(unsigned i=0; i<LineSize; i++) sum+=WriteBuf[LineIndex+i];
      if(!SendLine(hCom,str)) return false;
      LineCnt+=1;
      LineIndex+=LineSize;
      if(LineCnt>=20 || LineIndex>=WriteSize) {
        sprintf(str,"%d",sum);
        if(!SendLine(hCom,str)) return false;
        if(!WaitForResponse(hCom,"OK")) return false;
        LineCnt=sum=0;
      } // if
    } // while

    // prepare sector
    sector=adr/SectorSize;
    sprintf(str,"P %d %d",sector,sector);
    if(!SendLine(hCom,str)) return false;
    if(!WaitForResponse(hCom,CMD_SUCCESS)) return false;


    // copy data to flash memory
    sprintf(str,"C %d %d %d",adr,RamBufAdr,RamBufSize);
    if(!SendLine(hCom,str)) return false;
    if(!WaitForResponse(hCom,CMD_SUCCESS)) return false;

    adr+=WriteSize;
  } // while

  return true;
} // WriteFlash

bool ExitIspMode(HANDLE hCom)
{
    // disable BOOT
    if(!EscapeCommFunction(hCom,CLRRTS)) return false;

    // enable RESET
    if(!EscapeCommFunction(hCom,SETDTR)) return false;

    Sleep(50);

    // desable RESET
    if(!EscapeCommFunction(hCom,CLRDTR)) return false;

    return true;
} // ExitIspMode

AnsiString ProcessFile(AnsiString FileName, int ComNo, bool DeleteFileAfterWrite, void (*UpdateProgressBar)(int))
{
  AnsiString result="";
  FILE *f;
  fpos_t FileSize;
  HANDLE hCom;
  DCB dcb;
  COMMTIMEOUTS ct;
  char str[100];
  int PartId;
  int SectorNum;

  try {  // catch exception
    f=fopen(FileName.c_str(),"rb");
    if(f==NULL) throw "t@CJ܂ł";

    try { // guard f
      // get file size
      if(fseek(f,0,SEEK_END)!=0 || fgetpos(f,&FileSize)!=0 || fseek(f,0,SEEK_SET) !=0) throw "t@C̓ǂݍ݂Ɏs܂B";
      if(FileSize==0) throw ".bint@C̃TCY0łB߂܂B";
      if(FileSize>FlashSize) throw ".bint@C̃TCY傫܂B߂܂B";

      // open serial port
      sprintf(str,"\\\\.\\COM%d",ComNo);
      hCom=CreateFile(str,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
      if(hCom==INVALID_HANDLE_VALUE) throw "COM|[gJ܂B";

      try { // guard hCom
        memset(&dcb,0,sizeof(DCB));
        dcb.DCBlength        =sizeof(DCB);
        dcb.BaudRate         =CBR_9600;
        dcb.fBinary          =TRUE;
        dcb.fParity          =FALSE;
        dcb.fOutxCtsFlow     =FALSE;
        dcb.fOutxDsrFlow     =FALSE;
        dcb.fDtrControl      =DTR_CONTROL_DISABLE;
        dcb.fDsrSensitivity  =FALSE;
        dcb.fTXContinueOnXoff=TRUE;
        dcb.fOutX            =TRUE;
        dcb.fInX             =TRUE;
        dcb.fErrorChar       =FALSE;
        dcb.fNull            =FALSE;
        dcb.fRtsControl      =RTS_CONTROL_DISABLE;
        dcb.fAbortOnError    =FALSE;
//      dcb.fDummy2          =0;
//      dcb.wReserved        =0;
        dcb.XonLim           =10;
        dcb.XoffLim          =10;
        dcb.ByteSize         =8;
        dcb.Parity           =NOPARITY;
        dcb.StopBits         =ONESTOPBIT;
        dcb.XonChar          ='\x11';
        dcb.XoffChar         ='\x13';
//      dcb.ErrorChar        ='\x00';
//      dcb.EofChar          ='\x00';
//      dcb.EvtChar          ='\x00';

        if(!SetCommState(hCom,&dcb)) throw 0;

        ct.ReadIntervalTimeout        =   0;
        ct.ReadTotalTimeoutMultiplier =   0;
        ct.ReadTotalTimeoutConstant   =2000;
        ct.WriteTotalTimeoutMultiplier=   0;
        ct.WriteTotalTimeoutConstant  = 500;

        if(!SetCommTimeouts(hCom,&ct)) throw 0;

        if(!EnterIspMode(hCom) || !KillEcho(hCom)) throw "ISP[hɓ܂łB";
        if(!SetBaudRate(hCom,57600,1)) throw 1;

        dcb.BaudRate=CBR_57600;
        if(!SetCommState(hCom,&dcb)) throw 1;

        ReadPartId(hCom,str,sizeof(str));
        PartId=atoi(str);
        if(PartId!=0x0a40902b && PartId!=0x1a40902b) throw "wȊÕ}CRĂ܂B";

        if(!unlock(hCom)) throw 2;

        SectorNum=(FileSize+SectorSize-1)/SectorSize;
        if(!EraseFlash(hCom,SectorNum)) throw 2;
        if(!WriteFlash(hCom,f,FileSize,UpdateProgressBar)) {
          if(CancelBtnPressed) {
            throw CancelMsg;
          } else {
            throw "vȌ݂Ɏs܂B";
          }
        }
        if(!ExitIspMode(hCom)) throw "ISP[h甲̂Ɏs܂B";
      } // try
        __finally {
          CloseHandle(hCom);
      } // __finally
    } // try
    __finally {
      fclose(f);
    } // __finally

    if(DeleteFileAfterWrite) {
      if(remove(FileName.c_str())!=0) {
        throw ".bint@Cł܂łB";
      } // if
    } // if
  } // try
  catch(const char *e) {
    result=e;
  } // catch
  catch(int e) {
    switch(e) {
    case 0:
      result="ʐMɃG[܂";
      break;
    case 1:
      result="{[[g؂ւɎs܂B";
      break;
    case 2:
      result="tbV̏Ɏs܂B";
      break;
    } // switch
  } // catch

  return result;
} // ProcessFile

void UpdateProgressBar(int position)
{
  MainForm->WriteProgressBar->Position=position;
  Application->ProcessMessages();
} // TMainForm::UpdateProgressBar

void TMainForm::SetMainStatusBarIdle()
{
  if(AutoWriteChkBx->Enabled && AutoWriteChkBx->Checked) {
    MainStatusBar->SimpleText="w.bint@C܂̂҂Ă܂B";
  } else {
    MainStatusBar->SimpleText="COMԍƁA.bint@CIŁAuށv{^NbNĂB";
  } // if
} // TMainForm::SetMainStatusBarIdle

void TMainForm::EnableComponents()
{
  AutoWriteChkBx->Enabled=!busy && DeleteChkBx->Checked && FileNameEdt->Text!="";
  ComEdt->Enabled=!busy && (!AutoWriteChkBx->Enabled || !AutoWriteChkBx->Checked);
  FileNameEdt->Enabled=ComEdt->Enabled;
  FileNameBtn->Enabled=ComEdt->Enabled;
  WriteBtn->Enabled=busy || ((!AutoWriteChkBx->Enabled || !AutoWriteChkBx->Checked) && FileNameEdt->Text!="");
  DeleteChkBx->Enabled=!busy;
} // EnableComponents()

//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::WriteBtnClick(TObject *Sender)
{
  int ComNo;
  AnsiString result;

  if(busy) {
    CancelBtnPressed=true;
    return;
  } // if

  busy=true;
  CancelBtnPressed=false;
  WriteBtn->Caption="ݒ~";
  Screen->Cursor=crHourGlass;
  MainStatusBar->SimpleText="vOł܂B";
  WriteProgressBar->Position=0;
  WriteProgressBar->Visible=true;
  EnableComponents();
  Update();
  try { // busy guard WriteBtn, cursor, status bar and progress bar
    try { // catch exeption
      ComNo=StrToInt(ComEdt->Text);
      result=ProcessFile(FileNameEdt->Text,ComNo,DeleteChkBx->Checked,UpdateProgressBar);
    } // try
    catch(EConvertError *) {
      result="COMԍsł";
    } // catch
  } // try
  __finally {
    busy=false;
    EnableComponents();
    WriteBtn->Caption="";
    Screen->Cursor=crArrow;
    SetMainStatusBarIdle();
    WriteProgressBar->Visible=false;
  } // __finally
  if(result=="") {
    MessageBeep(MB_ICONINFORMATION);
  } else {
    MessageBox(Application->Handle,result.c_str(),Caption.c_str(),(result==CancelMsg ? MB_ICONINFORMATION : MB_ICONSTOP)|MB_OK);
    AutoWriteChkBx->Checked=false;
  } // if
  EnableComponents();
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::ComEdtChange(TObject *Sender)
{
  try {
    StrToInt(ComEdt->Text);
    ComEdt->Font->Color=clWindowText;
  } // try
  catch(...) {
    ComEdt->Font->Color=clRed;
  } // __finally
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FileNameBtnClick(TObject *Sender)
{
  if(FileOpenDlg->Execute()) {
    FileNameEdt->Text=FileOpenDlg->FileName;
  } // if
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormShow(TObject *Sender)
{
  EnableComponents();
  SetMainStatusBarIdle();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AutoWriteChkBxClick(TObject *Sender)
{
  FileTimer->Enabled=AutoWriteChkBx->Checked && AutoWriteChkBx->Enabled;
  EnableComponents();
  SetMainStatusBarIdle();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DeleteChkBxClick(TObject *Sender)
{
  if(DeleteChkBx->Checked) AutoWriteChkBx->Checked=false;
  AutoWriteChkBxClick(NULL);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FileTimerTimer(TObject *Sender)
{
  FileTimer->Enabled=false;
  try {
    if(!FileExists(FileNameEdt->Text)) return;
    FormStyle=fsStayOnTop;
    WriteBtnClick(NULL);
    FormStyle=fsNormal;
  } // try
  __finally {
    FileTimer->Enabled=AutoWriteChkBx->Checked;
  } // __finally
}
//---------------------------------------------------------------------------



void __fastcall TMainForm::FileNameEdtChange(TObject *Sender)
{
  EnableComponents();
}
//---------------------------------------------------------------------------


